
/*

-- you may need a -I<path-to-catch2>

    /usr/bin/c++ -std=gnu++17 -g -DDEBUG -D_DEBUG -D_GLIBCXX_ASSERTIONS -g -O0 -o test.o -c test.cpp

 */

#define CATCH_CONFIG_RUNNER

#define CATCH_CONFIG_PREFIX_ALL
#include    <catch2/catch_session.hpp>
#include    <catch2/catch_test_macros.hpp>
#include    <catch2/catch_approx.hpp>
#include    <catch2/matchers/catch_matchers.hpp>

#include    <cmath>
#include    <exception>


class timespec_ex
    : public timespec
{
public:
    timespec_ex()
    {
        set(0L);
    }
    timespec_ex(timespec_ex const & t)
    {
        set(t);
    }
    timespec_ex(timespec const & t)
    {
        set(t);
    }
    timespec_ex(time_t sec, long nsec)
    {
        set(sec, nsec);
    }
    timespec_ex(std::int64_t nsec)
    {
        set(nsec);
    }
    // uncomment and it compiles, comment and it fails
    //timespec_ex(int sec)
    //{
    //    set(sec * static_cast<std::int64_t>(1'000'000'000));
    //}
    timespec_ex(double sec)
    {
        set(sec);
    }
    timespec_ex & operator = (timespec_ex t)
    {
        tv_sec = t.tv_sec;
        tv_nsec = t.tv_nsec;
        return *this;
    }
    timespec_ex & operator = (timespec const & t)
    {
        tv_sec = t.tv_sec;
        tv_nsec = t.tv_nsec;
        return *this;
    }
    timespec_ex & operator = (std::int64_t nsec)
    {
        set(nsec);
        return *this;
    }
    timespec_ex & operator = (double sec)
    {
        set(sec);
        return *this;
    }
    void set(timespec_ex const & t)
    {
        tv_sec = t.tv_sec;
        tv_nsec = t.tv_nsec;
    }
    void set(timespec const & t)
    {
        tv_sec = t.tv_sec;
        tv_nsec = t.tv_nsec;
    }
    void set(time_t sec, long nsec)
    {
        tv_sec = sec;
        tv_nsec = nsec;
    }
    void set(std::int64_t nsec)
    {
        bool const neg(nsec < 0);
        if(neg)
        {
            nsec = -nsec;
        }
        tv_sec = static_cast<time_t>(nsec / 1'000'000'000LL);
        tv_nsec = static_cast<long>(nsec % 1'000'000'000LL);
        if(neg)
        {
            *this = -*this;
        }
    }
    void set(double sec)
    {
        bool const neg(sec < 0.0);
        if(neg)
        {
            sec = -sec;
        }
        tv_sec = static_cast<time_t>(floor(sec));
        tv_nsec = static_cast<long>((sec - floor(sec)) * 1.0e9);
        if(neg)
        {
            *this = -*this;
        }
    }
    static timespec_ex gettime(clockid_t clk_id = CLOCK_REALTIME)
    {
        timespec_ex result;
        clock_gettime(clk_id, &result);
        return result;
    }
    std::int64_t to_nsec() const
    {
        return tv_nsec
             + tv_sec * 1'000'000'000LL;
    }
    std::int64_t to_usec() const
    {
        return tv_nsec / 1'000LL
             + tv_sec * 1'000'000LL;
    }
    double to_sec() const
    {
        return static_cast<double>(tv_sec)
             + static_cast<double>(tv_nsec) / 1.0e9;
    }
    std::string to_string(std::string const & format = std::string(), bool local = false) const
    {
        struct tm date_and_time = {};
        struct tm * ptr(nullptr);
        if(local)
        {
            ptr = localtime_r(&tv_sec, &date_and_time);
        }
        else
        {
            ptr = gmtime_r(&tv_sec, &date_and_time);
        }
        if(ptr == nullptr)
        {
            throw std::overflow_error("the specified number of seconds could not be transformed in a 'struct tm'.");
        }
        std::string f(format);
        if(f.empty())
        {
            // TODO: actually retrieve the locale() format and
            //       search for "%T" or "%S" and insert ".%N"
            //       right after either
            //
            f = "%c.%N";
        }
        std::string::size_type pos(f.find("%N"));
        if(pos != std::string::npos)
        {
            std::string n(std::to_string(tv_nsec));
            if(n.length() > 9)
            {
                throw std::overflow_error("tv_nsec is 1 billion or more, which is invalid.");
            }
            std::string const indent(9 - n.length(), '0');
            n = indent + n;
            f = f.substr(0, pos) + n + f.substr(pos + 2);
        }
        char buf[256];
        size_t const sz(strftime(buf, sizeof(buf), f.c_str(), &date_and_time));
        if(sz == 0)
        {
            // this happens with just a "%p" and "wrong locale"
            // or when the buffer is too small, which should not
            // be the case unless you add much more than the format
            // in that string
            //
            throw std::overflow_error(
                  "the specified strftime() format \""
                + format
                + "\" failed.");
        }

        return std::string(buf, sz);
    }
    bool valid() const
    {
        return tv_nsec < 1'000'000'000LL;
    }
    bool negative() const
    {
        return tv_sec < 0LL;
    }
    timespec_ex add(timespec_ex const & rhs) const
    {
        bool const lneg(negative());
        bool const rneg(rhs.negative());

        timespec_ex lp(lneg ? -*this : *this);
        timespec_ex rp(rneg ? -rhs : rhs);

        timespec_ex result;

        switch((lneg ? 1 : 0) + (rneg ? 2 : 0))
        {
        case 0:     // positive + positive
        case 3:     // negative + negative
            result.tv_sec = lp.tv_sec + rp.tv_sec;
            result.tv_nsec = lp.tv_nsec + rp.tv_nsec;
            break;

        case 1:     // negative + positive
            result.tv_sec = rp.tv_sec - lp.tv_sec;
            result.tv_nsec = rp.tv_nsec - lp.tv_nsec;
            break;

        case 2:     // positive + negative
            result.tv_sec = lp.tv_sec - rp.tv_sec;
            result.tv_nsec = lp.tv_nsec - rp.tv_nsec;
            break;

        }

        if(result.tv_nsec < 0)
        {
            --result.tv_sec;
            result.tv_nsec += 1'000'000'000L;
        }
        else if(result.tv_nsec >= 1'000'000'000)
        {
            ++result.tv_sec;
            result.tv_nsec -= 1'000'000'000;
        }

        if(lneg && rneg)
        {
            result = -result;
        }

        return result;
    }
    int compare(timespec_ex const & rhs) const
    {
        if(tv_sec == rhs.tv_sec)
        {
            return tv_nsec == rhs.tv_nsec
                    ? 0
                    : (tv_nsec < rhs.tv_nsec ? -1 : 1);
        }

        return tv_sec < rhs.tv_sec ? -1 : 1;
    }
    bool operator ! () const
    {
        return tv_sec == 0 && tv_nsec == 0;
    }
    timespec_ex & operator += (timespec_ex const & rhs)
    {
        *this = add(rhs);
        return *this;
    }
    timespec_ex & operator ++ ()
    {
        *this += 1L;
        return *this;
    }
    timespec_ex operator ++ (int)
    {
        timespec_ex result(*this);
        *this += 1L;
        return result;
    }
    timespec_ex operator + (timespec_ex const & t) const
    {
        timespec_ex result(*this);
        result += t;
        return result;
    }
    timespec_ex & operator -= (timespec_ex const & rhs)
    {
        *this = add(-rhs);
        return *this;
    }
    timespec_ex & operator -- ()
    {
        *this -= 1L;
        return *this;
    }
    timespec_ex operator -- (int)
    {
        timespec_ex result(*this);
        *this -= 1L;
        return result;
    }
    timespec_ex operator - () const
    {
        timespec_ex result(timespec{ -tv_sec, -tv_nsec });
        if(result.tv_nsec < 0)
        {
            --result.tv_sec;
            result.tv_nsec += 1'000'000'000L;
        }
        return result;
    }
    timespec_ex operator - (timespec_ex const & rhs) const
    {
        timespec_ex result(*this);
        result -= rhs;
        return result;
    }
    bool operator == (timespec_ex const & t) const
    {
        return compare(t) == 0;
    }
    bool operator != (timespec_ex const & t) const
    {
        return compare(t) != 0;
    }
    bool operator < (timespec_ex const & t) const
    {
        return compare(t) == -1;
    }
    bool operator <= (timespec_ex const & t) const
    {
        return compare(t) <= 0;
    }
    bool operator > (timespec_ex const & t) const
    {
        return compare(t) == 1;
    }
    bool operator >= (timespec_ex const & t) const
    {
        return compare(t) >= 0;
    }
};

CATCH_TEST_CASE("equal", "")
{
    timespec_ex a;
    timespec_ex c;

    // error occurs on the following line:
    CATCH_REQUIRE(a == c);
}


// vim: ts=4 sw=4 et
